//////////////////////////////////////////////////////////////////////////
// BonSDK Stream service
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.

using System;
using System.Collections.Generic;
using System.Net.Sockets;
using System.Runtime.ConstrainedExecution;
using System.Runtime.InteropServices;
using NetFwTypeLib;

namespace Bon
{
	internal sealed class FirewallManager : CriticalFinalizerObject, IDisposable
	{
		private static readonly Type firewallManagerType_ = Type.GetTypeFromProgID("HNetCfg.FwMgr", false);
		private static readonly Type openPortType_ = Type.GetTypeFromProgID("HNetCfg.FWOpenPort", false);

		private readonly string policyName_;
		private readonly bool scopeFree_;
		private readonly INetFwMgr firewallManager_;
		private readonly INetFwProfile profile_;
		private readonly List<KeyValuePair<NET_FW_IP_PROTOCOL_, int>> openPorts_ = new List<KeyValuePair<NET_FW_IP_PROTOCOL_, int>>();

		public FirewallManager(string policyName, bool scopeFree)
		{
			policyName_ = policyName;
			scopeFree_ = scopeFree;

			firewallManager_ = (INetFwMgr)Activator.CreateInstance(firewallManagerType_);
			profile_ = firewallManager_.LocalPolicy.CurrentProfile;
		}

		~FirewallManager()
		{
			Dispose();
		}

		public void Dispose()
		{
			lock (this)
			{
				foreach (KeyValuePair<NET_FW_IP_PROTOCOL_, int> definition in openPorts_)
				{
					profile_.GloballyOpenPorts.Remove(definition.Value, definition.Key);
				}
			}
			GC.SuppressFinalize(this);
		}

		public static bool HasWindowsFirewall
		{
			get
			{
				return (firewallManagerType_ != null) && (openPortType_ != null);
			}
		}

		public void Add(string name, ProtocolType protocol, int port)
		{
			lock (this)
			{
				NET_FW_IP_PROTOCOL_ rawProtocol = NET_FW_IP_PROTOCOL_.NET_FW_IP_PROTOCOL_ANY;
				switch (protocol)
				{
					case ProtocolType.Tcp:
						rawProtocol = NET_FW_IP_PROTOCOL_.NET_FW_IP_PROTOCOL_TCP;
						break;
					case ProtocolType.Udp:
						rawProtocol = NET_FW_IP_PROTOCOL_.NET_FW_IP_PROTOCOL_UDP;
						break;
					case ProtocolType.Unspecified:
						rawProtocol = NET_FW_IP_PROTOCOL_.NET_FW_IP_PROTOCOL_ANY;
						break;
					default:
						throw new ArgumentException("Cannot filter protocol: " + protocol.ToString());
				}

				string rawName = string.Format("{0} ({1})", policyName_, name);
				bool found = false;
				foreach (INetFwOpenPort openPort in profile_.GloballyOpenPorts)
				{
					if (openPort.Name == rawName)
					{
						openPort.Protocol = rawProtocol;
						openPort.Port = port;
						openPort.Scope = scopeFree_ ? NET_FW_SCOPE_.NET_FW_SCOPE_ALL : NET_FW_SCOPE_.NET_FW_SCOPE_LOCAL_SUBNET;
						openPort.Enabled = true;
						found = true;
						break;
					}
				}
				if (found == false)
				{
					INetFwOpenPort openPort = (INetFwOpenPort)Activator.CreateInstance(openPortType_);
					openPort.Name = rawName;
					openPort.Protocol = rawProtocol;
					openPort.Port = port;
					openPort.Scope = scopeFree_ ? NET_FW_SCOPE_.NET_FW_SCOPE_ALL : NET_FW_SCOPE_.NET_FW_SCOPE_LOCAL_SUBNET;
					openPort.Enabled = true;
					profile_.GloballyOpenPorts.Add(openPort);
				}

				openPorts_.Add(new KeyValuePair<NET_FW_IP_PROTOCOL_, int>(rawProtocol, port));
			}
		}
	}
}
